

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.3.210310 (10 March 2021)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2021
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#if !defined(_REQUESTEXECUTOR_H_)
#define _REQUESTEXECUTOR_H_

#pragma once

#include "NetworkManager.h"
#include "RequestClient.h"


namespace	cmlabs{

/////////////////////////////////////////////////////////////
// Request Executor
/////////////////////////////////////////////////////////////

class RequestExecutor : public Runnable, public NetworkReceiver {
public:
	friend THREAD_RET THREAD_FUNCTION_CALL RequestExecutorRun(THREAD_ARG arg);

	RequestExecutor(uint32 id = 0, const char* name = NULL);
	virtual ~RequestExecutor();
	bool shutdownNetwork();

	bool addLongRequestName(const char* name);
	bool addGateway(uint32 id, std::string addr, uint16 port, uint8 encryption = NOENC);
	bool setLongRequestLimit(uint32 limit);

	bool receiveNetworkEvent(NetworkEvent* evt, NetworkChannel* channel, uint64 conid);
	bool receiveMessage(DataMessage* msg, NetworkChannel* channel, uint64 conid);
	bool receiveHTTPRequest(HTTPRequest* req, NetworkChannel* channel, uint64 conid) {return false;}

	DataMessage* waitForLongRequest(uint32 timeoutMS);
	DataMessage* waitForShortRequest(uint32 timeoutMS);
	bool replyToQuery(DataMessage* msg);

protected:
	bool run();
	uint32 sendStatusNow();
	bool sendMessageToGateway(DataMessage* msg, RequestGatewayConnection& con);

	std::list<RequestGatewayConnection> connections;
	//std::map<uint64, RequestGatewayConnection> gateways;

	bool replyToGateway(DataMessage* msg);

	uint32 executorID;
	std::string executorName;
	NetworkManager* manager;
	NetworkChannel* channel;
	utils::WaitQueuePointer<DataMessage*> shortExecQ;
	utils::WaitQueuePointer<DataMessage*> longExecQ;
	utils::WaitQueuePointer<DataMessage*> replyQ;
	std::map<uint64,RequestReply*> requestMap;
	std::vector<std::string> longReqNames;
	utils::Mutex mutex;
	utils::Mutex conMutex;
	utils::Semaphore shortExecQSemaphore;
	utils::Semaphore longExecQSemaphore;
	uint32 threadID;
	uint64 lastRefID;
	uint64 repliedCount;
	uint64 sentCount;
	uint64 shortReceivedCount;
	uint64 longReceivedCount;
	int32 heartbeatIntervalMS;
	uint64 lastHeartbeat;
	uint32 longReqLimit;
	//MovingAverage shortAvgStats;
	//MovingAverage longAvgStats;
};

THREAD_RET THREAD_FUNCTION_CALL RequestExecutorRun(THREAD_ARG arg);

class TestRequestExecutor : public RequestExecutor {
public:
	friend THREAD_RET THREAD_FUNCTION_CALL ShortExecutorRun(THREAD_ARG arg);
	friend THREAD_RET THREAD_FUNCTION_CALL LongExecutorRun(THREAD_ARG arg);
	TestRequestExecutor(uint32 id, uint32 processTimeMS);
protected:
	bool shortRequestRun();
	bool longRequestRun();
	uint32 id;
	uint32 processTimeMS;
};

THREAD_RET THREAD_FUNCTION_CALL ShortExecutorRun(THREAD_ARG arg);
THREAD_RET THREAD_FUNCTION_CALL LongExecutorRun(THREAD_ARG arg);

}

#endif // _REQUESTEXECUTOR_H_
